
// Single processor, ad++ subscripting reduced to canonical form

#include "ad++.h"
#include <sys/time.h>

const int N = 100 ;
const int NITER = 500 ;

int wlo [] = {1, 1}, whi [] = {1, 1} ;  // ghost widths for `writeHalo'

int main(int argc, char** argv) {
  AdlibInit(argc, argv) ;

  Procs2 p(1, 1) ;
  ON(p) {
    BlockRange x(N, p.dim(0), LAY_EXTENDED, 1, 1),
               y(N, p.dim(1), LAY_EXTENDED, 1, 1) ;

    float h = 1.0 / (N - 1) ;

    // Initialization (unoptimized):
    //    sol = exact solution
    //    f   = forcing term
    //    u   = approximation solution

    Array2<float> u(x, y), f(x, y), sol(x, y) ;

    const float fac = 4.0 - 2.0 * (cos(h) + cos(3 * h)) ;

    Index i(x), j(y) ;
    OVERALL(i) {
      OVERALL(j) {
	sol(i, j) = sin(3 * (h * i) + h * j) ;
        if(i == 0 || i == N - 1 || j == 0 || j == N - 1)
	  u(i, j) = sol(i, j) ;
        else {
	  u(i, j) = 0.0 ;
	  f(i, j) = fac * sol(i, j) ;
        }
      } ALLOVER(j) ;
    } ALLOVER(i) ;

//    gprintf("initial u:\n") ;
//    aprintf("%7.3f%*N", u, N) ;

    // Main loop

    Index i1(x.triplet(1, N - 2)) ;

    clock() ;

    for(int iter = 0 ; iter < NITER ; iter++) {
      for(int parity = 0 ; parity < 2 ; parity++) {

//      writeHalo(u, wlo, whi) ;

        OVERALL(i1) {
          Index j1(y.triplet(1 + (1 + i1 + parity) % 2, N - 2, 2)) ;

          OVERALL(j1) {
            float* u_dat = u.dat() ;
            float* f_dat = f.dat() ;

            u_dat[u.str(0).offset(i1) + u.str(1).offset(j1)] =
                0.25 * (f_dat[f.str(0).offset(i1) + f.str(1).offset(j1)] +
                        u_dat[u.str(0).offset(i1 >> 1) + u.str(1).offset(j1)] +
                        u_dat[u.str(0).offset(i1 << 1) + u.str(1).offset(j1)] +
                        u_dat[u.str(0).offset(i1) + u.str(1).offset(j1 >> 1)] +
                        u_dat[u.str(0).offset(i1) + u.str(1).offset(j1 << 1)]) ;
          } ALLOVER(j1) ;
        } ALLOVER(i1) ;
      }

//      gprintf("iteration = %d\n", iter) ;
//      aprintf("%7.3f%*N", u, N) ;
    }

    clock_t t = clock() ;
  
    printf("elapsed time = %f\n", (float) t / CLOCKS_PER_SEC) ;

//    gprintf("exact solution:\n") ;
//    aprintf("%7.3f%*N", sol, N) ;

  } NO(p) ;

  AdlibFinalize() ;
}

